from __future__ import annotations

import json
import pathlib
import pickle
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pandas as pd
from psychopy.data import TrialHandlerExt
from psychopy.misc import fromFile
from vstt.display import default_display_options
from vstt.display import import_display_options
from vstt.meta import default_metadata
from vstt.meta import import_metadata
from vstt.stats import append_stats_data_to_excel
from vstt.stats import stats_dataframe
from vstt.trial import default_trial
from vstt.trial import import_and_validate_trial


class Experiment:
    def __init__(self, filename: Optional[str] = None):
        self.filename = "default-experiment.psydat"
        self.has_unsaved_changes = False
        self.metadata = default_metadata()
        self.display_options = default_display_options()
        self.trial_list = [default_trial()]
        self.trial_handler_with_results: Optional[TrialHandlerExt] = None
        self.stats: Optional[pd.DataFrame] = None
        if filename is not None:
            self.load_file(filename)

    def create_trialhandler(self) -> TrialHandlerExt:
        for index, trial in enumerate(self.trial_list):
            self.trial_list[index] = import_and_validate_trial(trial)
        return TrialHandlerExt(
            self.trial_list,
            nReps=1,
            method="sequential",
            originPath=-1,
            extraInfo={
                "display_options": self.display_options,
                "metadata": self.metadata,
            },
        )

    def clear_results(self) -> None:
        self.trial_handler_with_results = None

    def _as_dict(self) -> Dict[str, Any]:
        return {
            "metadata": self.metadata,
            "display_options": self.display_options,
            "trial_list": self.trial_list,
        }

    def load_file(self, filename: str) -> None:
        suffix = pathlib.Path(filename).suffix
        if suffix == ".xlsx":
            self.load_excel(filename)
        elif suffix == ".json":
            self.load_json(filename)
        else:
            # assume psydat format by default
            self.load_psydat(filename)

    def save_psydat(self, filename: str) -> None:
        if not filename.endswith(".psydat"):
            filename += ".psydat"

        if self.trial_handler_with_results is not None:
            # save existing trial handler with results, only update its extraInfo dict
            self.trial_handler_with_results.extraInfo = {
                "display_options": self.display_options,
                "metadata": self.metadata,
            }
            self.trial_handler_with_results.saveAsPickle(
                filename, fileCollisionMethod="overwrite"
            )
        else:
            # create a new trial handler to save this experiment
            trial_handler = self.create_trialhandler()
            # temporary hack as the psyschopy function only saves if there are results!
            with open(filename, "wb") as f:
                pickle.dump(trial_handler, f)
        self.filename = filename
        self.has_unsaved_changes = False

    def load_psydat(self, filename: str) -> None:
        self.import_and_validate_trial_handler(fromFile(filename))
        self.filename = filename
        self.has_unsaved_changes = False

    def save_excel(self, filename: str, data_format: str) -> None:
        """
        data_format can be
        - "trial": one sheet of data exported per trial
        - "target: one sheet of data exported per target
        """
        if not filename.endswith(".xlsx"):
            filename += ".xlsx"
        with pd.ExcelWriter(filename) as writer:
            pd.DataFrame([self.metadata]).to_excel(
                writer, sheet_name="metadata", index=False
            )
            pd.DataFrame([self.display_options]).to_excel(
                writer, sheet_name="display_options", index=False
            )
            pd.DataFrame(self.trial_list).to_excel(
                writer, sheet_name="trial_list", index=False
            )
            if self.stats is not None:
                append_stats_data_to_excel(self.stats, writer, data_format)

    def load_excel(self, filename: str) -> None:
        dfs = pd.read_excel(filename, ["metadata", "display_options", "trial_list"])
        self.import_and_validate_dicts(
            filename,
            dfs["metadata"].to_dict("records")[0],
            dfs["display_options"].to_dict("records")[0],
            dfs["trial_list"].to_dict("records"),
        )

    def save_json(self, filename: str) -> None:
        if not filename.endswith(".json"):
            filename += ".json"
        with open(filename, "w") as f:
            json.dump(self._as_dict(), f)

    def load_json(self, filename: str) -> None:
        with open(filename, "r") as f:
            d = json.load(f)
        self.import_and_validate_dicts(
            filename, d["metadata"], d["display_options"], d["trial_list"]
        )

    def import_and_validate_dicts(
        self,
        filename: str,
        metadata_dict: Dict,
        display_options_dict: Dict,
        trial_dict_list: List[Dict],
    ) -> None:
        self.metadata = import_metadata(metadata_dict)
        self.display_options = import_display_options(display_options_dict)
        self.trial_list = [
            import_and_validate_trial(trial_dict) for trial_dict in trial_dict_list
        ]
        self.trial_handler_with_results = None
        self.stats = None
        self.has_unsaved_changes = True
        self.filename = str(pathlib.Path(filename).with_suffix(".psydat"))

    def import_and_validate_trial_handler(self, trial_handler: TrialHandlerExt) -> None:
        # psychopy trial handler converts empty trial list [] -> [None]
        if trial_handler.trialList == [None]:
            trial_handler.trialList = []
        for index, trial in enumerate(trial_handler.trialList):
            trial_handler.trialList[index] = import_and_validate_trial(trial)
        self.trial_list = trial_handler.trialList
        if not trial_handler.extraInfo:
            trial_handler.extraInfo = {}
        self.metadata = import_metadata(
            trial_handler.extraInfo.get("metadata", default_metadata())
        )
        self.display_options = import_display_options(
            trial_handler.extraInfo.get("display_options", default_display_options())
        )
        if trial_handler.finished:
            self.trial_handler_with_results = trial_handler
            self.stats = stats_dataframe(trial_handler)
        else:
            self.trial_handler_with_results = None
            self.stats = None



from __future__ import annotations

from typing import Callable
from typing import Dict
from typing import Optional

from psychopy.visual.window import Window
from PyQt5 import QtCore
from PyQt5 import QtWidgets
from vstt.display import display_options_labels
from vstt.experiment import Experiment


class DisplayOptionsWidget(QtWidgets.QWidget):
    experiment_modified = QtCore.pyqtSignal()

    def __init__(
        self, parent: Optional[QtWidgets.QWidget] = None, win: Optional[Window] = None
    ):
        super().__init__(parent)
        self._win = win
        self._experiment: Experiment = Experiment()
        self._widgets: Dict[str, QtWidgets.QCheckBox] = {}

        outer_layout = QtWidgets.QVBoxLayout()
        group_box = QtWidgets.QGroupBox("Display Options")
        outer_layout.addWidget(group_box)
        inner_layout = QtWidgets.QVBoxLayout()
        group_box.setLayout(inner_layout)
        labels = display_options_labels()
        for row_index, key in enumerate(labels.keys()):
            checkbox = QtWidgets.QCheckBox(f"{labels[key]}", self)
            inner_layout.addWidget(checkbox)
            checkbox.clicked.connect(self._update_value_callback(key))
            self._widgets[key] = checkbox
        self.setLayout(outer_layout)

    def _update_value_callback(self, key: str) -> Callable[[bool], None]:
        def _update_value(value: bool) -> None:
            self._experiment.display_options[key] = value  # type: ignore
            self._experiment.has_unsaved_changes = True
            self.experiment_modified.emit()

        return _update_value

    @property
    def experiment(self) -> Experiment:
        return self._experiment

    @experiment.setter
    def experiment(self, experiment: Experiment) -> None:
        self._experiment = experiment
        for key, widget in self._widgets.items():
            widget.setChecked(self._experiment.display_options[key])  # type: ignore



from __future__ import annotations

import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional

import numpy as np
from psychopy.gui import DlgFromDict
from vstt.common import import_typed_dict
from vstt.vtypes import Trial


def describe_trial(trial: Trial) -> str:
    repeats = f"{trial['weight']} repeat{'s' if trial['weight'] > 1 else ''}"
    targets = f"target{'s' if trial['num_targets'] > 1 else ''}"
    if trial["condition_timeout"] > 0.0:
        repeats = f"up to {repeats}"
        targets = f"{targets} in {trial['condition_timeout']}s"
    return f"{repeats} of {trial['num_targets']} {trial['target_order']} {targets}"


def describe_trials(trials: List[Trial]) -> str:
    return "\n".join(["  - " + describe_trial(trial) for trial in trials])


def default_trial() -> Trial:
    return {
        "weight": 1,
        "condition_timeout": 0.0,
        "num_targets": 8,
        "target_order": "clockwise",
        "target_indices": "0 1 2 3 4 5 6 7",
        "add_central_target": True,
        "hide_target_when_reached": True,
        "show_target_labels": False,
        "target_labels": "0 1 2 3 4 5 6 7",
        "fixed_target_intervals": False,
        "target_duration": 5.0,
        "central_target_duration": 5.0,
        "pre_target_delay": 0.0,
        "pre_central_target_delay": 0.0,
        "pre_first_target_extra_delay": 0.0,
        "target_distance": 0.4,
        "target_size": 0.04,
        "central_target_size": 0.02,
        "show_inactive_targets": True,
        "ignore_incorrect_targets": True,
        "play_sound": True,
        "use_joystick": False,
        "joystick_max_speed": 0.02,
        "show_cursor": True,
        "cursor_size": 0.02,
        "show_cursor_path": True,
        "automove_cursor_to_center": False,
        "freeze_cursor_between_targets": False,
        "cursor_rotation_degrees": 0.0,
        "post_trial_delay": 0.0,
        "post_trial_display_results": False,
        "post_block_delay": 10.0,
        "post_block_display_results": True,
        "show_delay_countdown": True,
        "enter_to_skip_delay": True,
    }


def trial_labels() -> Dict:
    return {
        "weight": "Repetitions",
        "condition_timeout": "Maximum time, 0=unlimited (secs)",
        "num_targets": "Number of targets",
        "target_order": "Target order",
        "target_indices": "Target indices",
        "add_central_target": "Add a central target",
        "hide_target_when_reached": "Hide target when reached",
        "show_target_labels": "Display target labels",
        "target_labels": "Target labels",
        "fixed_target_intervals": "Fixed target display intervals",
        "target_duration": "Target display duration (secs)",
        "central_target_duration": "Central target display duration (secs)",
        "pre_target_delay": "Delay before outer target display (secs)",
        "pre_central_target_delay": "Delay before central target display (secs)",
        "pre_first_target_extra_delay": "Extra delay before first outer target of condition (secs)",
        "target_distance": "Distance to targets (screen height fraction)",
        "target_size": "Target size (screen height fraction)",
        "central_target_size": "Central target size (screen height fraction)",
        "show_inactive_targets": "Show inactive targets",
        "ignore_incorrect_targets": "Ignore cursor hitting incorrect target",
        "play_sound": "Play a sound on target display",
        "use_joystick": "Control the cursor using a joystick",
        "joystick_max_speed": "Maximum joystick speed (screen height fraction per frame)",
        "show_cursor": "Show cursor",
        "cursor_size": "Cursor size (screen height fraction)",
        "show_cursor_path": "Show cursor path",
        "automove_cursor_to_center": "Automatically move cursor to center",
        "freeze_cursor_between_targets": "Freeze cursor until target is displayed",
        "cursor_rotation_degrees": "Cursor rotation (degrees)",
        "post_trial_delay": "Delay between trials (secs)",
        "post_trial_display_results": "Display results after each trial",
        "post_block_delay": "Delay after last trial (secs)",
        "post_block_display_results": "Display combined results after last trial",
        "show_delay_countdown": "Display a countdown during delays",
        "enter_to_skip_delay": "Skip delay by pressing enter key",
    }


def get_trial_from_user(
    initial_trial: Optional[Trial] = None,
) -> Optional[Trial]:
    if initial_trial:
        trial = copy.deepcopy(initial_trial)
    else:
        trial = default_trial()
    order_of_targets = [trial["target_order"]]
    for target_order in ["clockwise", "anti-clockwise", "random", "fixed"]:
        if target_order != order_of_targets[0]:
            order_of_targets.append(target_order)
    trial["target_order"] = order_of_targets
    dialog = DlgFromDict(
        trial, title="Trial conditions", labels=trial_labels(), sortKeys=False
    )
    if not dialog.OK:
        return None
    return import_and_validate_trial(trial)


def import_and_validate_trial(trial_or_dict: Mapping[str, Any]) -> Trial:
    trial = import_typed_dict(trial_or_dict, default_trial())
    # make any negative time durations zero
    for duration in [
        "condition_timeout",
        "target_duration",
        "central_target_duration",
        "pre_target_delay",
        "pre_central_target_delay",
        "pre_first_target_extra_delay",
        "post_trial_delay",
        "post_block_delay",
    ]:
        if trial[duration] < 0.0:  # type: ignore
            trial[duration] = 0.0  # type: ignore
    # convert string of target indices to a numpy array of ints
    target_indices = np.fromstring(trial["target_indices"], dtype="int", sep=" ")
    if trial["target_order"] == "fixed":
        # clip indices to valid range
        target_indices = np.clip(target_indices, 0, trial["num_targets"] - 1)
    else:
        # construct clockwise sequence
        target_indices = np.array(range(trial["num_targets"]))
        if trial["target_order"] == "anti-clockwise":
            target_indices = np.flip(target_indices)
        elif trial["target_order"] == "random":
            rng = np.random.default_rng()
            rng.shuffle(target_indices)
    # convert array of indices back to a string
    trial["target_indices"] = " ".join(f"{int(i)}" for i in target_indices)
    return trial



from __future__ import annotations

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from psychopy.data import TrialHandlerExt
from psychopy.event import xydist


def list_dest_stat_label_units() -> List[Tuple[str, List[Tuple[str, str, str]]]]:
    list_dest_stats = []
    for destination in ["target", "center"]:
        stats = []
        for base_stat, label, unit in [
            ("reaction_time", "Reaction", "s"),
            ("movement_time", "Movement", "s"),
            ("time", "Time", "s"),
            ("distance", "Distance", ""),
            ("rmse", "RMSE", ""),
            ("success", "Success", ""),
        ]:
            stats.append((f"to_{destination}_{base_stat}", label, unit))
        list_dest_stats.append((destination, stats))
    return list_dest_stats


def _get_trial_data_columns() -> List[str]:
    return [
        "i_trial",
        "i_rep",
        "i_target",
        "condition_index",
        "target_index",
        "target_pos",
        "to_target_timestamps",
        "to_target_mouse_positions",
        "to_target_success",
        "to_target_num_timestamps_before_visible",
        "center_pos",
        "to_center_timestamps",
        "to_center_mouse_positions",
        "to_center_success",
        "to_center_num_timestamps_before_visible",
    ]


def _get_dat(
    data: Dict, key: str, index: Tuple, i_target: int, default_value: Any
) -> Any:
    ar = data.get(key)
    if ar is None:
        logging.debug(
            f"Key '{key}' not found in data, using default value {default_value}"
        )
        return default_value
    try:
        return ar[index][i_target]
    except IndexError:
        logging.debug(
            f"Index error for key '{key}', index '{index}', i_target '{i_target}', using default value {default_value}"
        )
    return default_value


def _get_target_data(
    trial_handler: TrialHandlerExt, index: Tuple, i_target: int
) -> List:
    data = trial_handler.data
    condition_index = trial_handler.sequenceIndices[index]
    target_index = data["target_indices"][index][i_target]
    target_pos = np.array(data["target_pos"][index][i_target])
    center_pos = np.array([0.0, 0.0])
    to_target_timestamps = np.array(data["to_target_timestamps"][index][i_target])
    to_target_num_timestamps_before_visible = _get_dat(
        data, "to_target_num_timestamps_before_visible", index, i_target, 0
    )
    to_target_mouse_positions = np.stack(
        np.array(data["to_target_mouse_positions"][index][i_target])
    )  # type: ignore
    to_target_success = np.array(data["to_target_success"][index][i_target])
    to_center_success: Union[np.ndarray, bool]
    to_center_timestamps = np.array(
        _get_dat(data, "to_center_timestamps", index, i_target, [])
    )
    to_center_num_timestamps_before_visible = _get_dat(
        data, "to_center_num_timestamps_before_visible", index, i_target, 0
    )
    to_center_mouse_positions = np.array(
        _get_dat(data, "to_center_mouse_positions", index, i_target, [])
    )
    if to_center_mouse_positions.shape[0] > 0:
        to_center_mouse_positions = np.stack(to_center_mouse_positions)  # type: ignore
    to_center_success = np.array(
        _get_dat(data, "to_center_success", index, i_target, True)
    )
    return [
        index[0],
        index[1],
        i_target,
        condition_index,
        target_index,
        target_pos,
        to_target_timestamps,
        to_target_mouse_positions,
        to_target_success,
        to_target_num_timestamps_before_visible,
        center_pos,
        to_center_timestamps,
        to_center_mouse_positions,
        to_center_success,
        to_center_num_timestamps_before_visible,
    ]


def _data_df(trial_handler: TrialHandlerExt) -> pd.DataFrame:
    data = []
    for index in np.ndindex(trial_handler.sequenceIndices.shape):
        # trials that have not yet happened have a default string instead of an array in their data
        if type(trial_handler.data["target_indices"][index]) is np.ndarray:
            n_targets = trial_handler.data["target_indices"][index].shape[0]
            for i_target in range(n_targets):
                data.append(_get_target_data(trial_handler, index, i_target))
    return pd.DataFrame(data, columns=_get_trial_data_columns())


def stats_dataframe(trial_handler: TrialHandlerExt) -> pd.DataFrame:
    df = _data_df(trial_handler)
    for destination in ["target", "center"]:
        df[f"to_{destination}_distance"] = df[f"to_{destination}_mouse_positions"].map(
            _distance
        )
        df[f"to_{destination}_reaction_time"] = df.apply(
            lambda x: _reaction_time(
                x[f"to_{destination}_timestamps"],
                x[f"to_{destination}_mouse_positions"],
                x[f"to_{destination}_num_timestamps_before_visible"],
            ),
            axis=1,
        )
        df[f"to_{destination}_time"] = df.apply(
            lambda x: _total_time(
                x[f"to_{destination}_timestamps"],
                x[f"to_{destination}_num_timestamps_before_visible"],
            ),
            axis=1,
        )
        df[f"to_{destination}_movement_time"] = (
            df[f"to_{destination}_time"] - df[f"to_{destination}_reaction_time"]
        )
        df[f"to_{destination}_rmse"] = df.apply(
            lambda x: _rmse(
                x[f"to_{destination}_mouse_positions"], x[f"{destination}_pos"]
            ),
            axis=1,
        )
    return df


def append_stats_data_to_excel(df: pd.DataFrame, writer: Any, data_format: str) -> None:
    """
    data_format can be
    - "trial": one sheet of data exported per trial
    - "target: one sheet of data exported per target
    """
    if data_format not in ["trial", "target"]:
        raise RuntimeError(f"data_format '{data_format}' not supported")
    data_labels = [
        "to_target_timestamps",
        "to_target_mouse_positions",
        "to_center_timestamps",
        "to_center_mouse_positions",
    ]
    # first sheet: all statistics, exclude arrays of timestamp/position data
    # convert columns of xy values to x column and y column of floats
    df_stats = df.drop(columns=data_labels)
    for label in ["target_pos", "center_pos"]:
        column_as_2d_array = np.stack(df_stats[f"{label}"].to_numpy())
        df_stats[f"{label}_x"] = column_as_2d_array[:, 0]
        df_stats[f"{label}_y"] = column_as_2d_array[:, 1]
        df_stats = df_stats.drop(columns=[label])
    df_stats.to_excel(writer, sheet_name="statistics", index=False)
    # add timestamp/mouse position arrays
    if data_format == "target":
        # one sheet for each row (target) in df, with arrays of time/position data.
        # arrays transposed to columns, x-y pairs are split into two columns.
        # 6 columns:
        #   - to_target_timestamps
        #   - to_center_timestamps
        #   - to_target_mouse_positions_x
        #   - to_target_mouse_positions_y
        #   - to_center_mouse_positions_x
        #   - to_center_mouse_positions_y
        for row in df.itertuples():
            df_data = pd.concat(
                [
                    pd.DataFrame({label: getattr(row, label)})
                    for label in ["to_target_timestamps", "to_center_timestamps"]
                ],
                axis=1,
            )
            for label in ["to_target_mouse_positions", "to_center_mouse_positions"]:
                arr = getattr(row, label)
                if arr.shape[-1] > 0:
                    df_data = pd.concat(
                        [
                            df_data,
                            pd.DataFrame(
                                {f"{label}_x": arr[:, 0], f"{label}_y": arr[:, 1]}
                            ),
                        ],
                        axis=1,
                    )
            df_data.to_excel(writer, sheet_name=f"{row.Index}", index=False)
    else:
        # one sheet per trial: all targets for a trial are concatenated
        # 6 columns:
        #   - timestamps (start from zero, increases throughout the trial)
        #   - mouse_x
        #   - mouse_y
        #   - target_index (-99 if no target is currently displayed)
        #   - target_x (-99 if no target is currently displayed)
        #   - target_y (-99 if no target is currently displayed)
        for i_trial in df.i_trial.unique():
            times = np.array([])
            x_positions = np.array([])
            y_positions = np.array([])
            i_targets = np.array([], dtype=np.int64)
            x_targets = np.array([])
            y_targets = np.array([])
            for row in df.loc[df.i_trial == i_trial].itertuples():
                for dest in ["target", "center"]:
                    raw_times = getattr(row, f"to_{dest}_timestamps")
                    if raw_times.shape[-1] > 0:
                        times = np.append(times, raw_times)
                        points = getattr(row, f"to_{dest}_mouse_positions")
                        x_positions = np.append(x_positions, points[:, 0])
                        y_positions = np.append(y_positions, points[:, 1])
                        # set target info for all timestamps
                        if dest == "center":
                            pos = (0.0, 0.0)
                            i_target = -1  # give center target the special index -1
                            num_timestamps_before_target_visible = (
                                row.to_center_num_timestamps_before_visible
                            )
                        else:
                            pos = row.target_pos
                            i_target = row.i_target
                            num_timestamps_before_target_visible = (
                                row.to_target_num_timestamps_before_visible
                            )
                        target_i = np.full_like(raw_times, i_target, dtype=np.int64)
                        target_x = np.full_like(raw_times, pos[0])
                        target_y = np.full_like(raw_times, pos[1])
                        # if no target is visible use the special index -99
                        target_i[0:num_timestamps_before_target_visible] = -99
                        target_x[0:num_timestamps_before_target_visible] = -99
                        target_y[0:num_timestamps_before_target_visible] = -99
                        i_targets = np.append(i_targets, target_i)
                        x_targets = np.append(x_targets, target_x)
                        y_targets = np.append(y_targets, target_y)
            df_data = pd.DataFrame(
                {
                    "timestamps": times,
                    "mouse_positions_x": x_positions,
                    "mouse_positions_y": y_positions,
                    "i_target": i_targets,
                    "target_x": x_targets,
                    "target_y": y_targets,
                }
            )
            df_data.to_excel(writer, sheet_name=f"{i_trial}", index=False)


def _reaction_time(
    mouse_times: np.ndarray,
    mouse_positions: np.ndarray,
    to_target_num_timestamps_before_visible: int,
    epsilon: float = 1e-12,
) -> float:
    """
    The reaction time is defined as the timestamp where the cursor first moves,
    minus the timestamp where the target becomes visible.
    This means the reaction time can be negative if the cursor is moved before
    the target is made visible.

    :param mouse_times: The array of timestamps
    :param mouse_positions: The array of mouse positions
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :param epsilon: The minimum euclidean distance to qualify as moving the cursor
    :return: The reaction time
    """
    if (
        mouse_times.shape[0] != mouse_positions.shape[0]
        or mouse_times.shape[0] == 0
        or mouse_times.shape[0] < to_target_num_timestamps_before_visible
    ):
        return np.nan
    i = 0
    while xydist(mouse_positions[0], mouse_positions[i]) < epsilon and i + 1 < len(
        mouse_times
    ):
        i += 1
    return mouse_times[i] - mouse_times[to_target_num_timestamps_before_visible]


def _total_time(
    mouse_times: np.ndarray,
    to_target_num_timestamps_before_visible: int,
) -> float:
    """
    The time to target is defined as the final timestamp (corresponding either to the target being reached
    or a timeout) minus the timestamp where the target becomes visible.

    :param mouse_times: The array of timestamps
    :param to_target_num_timestamps_before_visible: The index of the first timestamp where the target is visible
    :return: The total time to target
    """
    if (
        mouse_times.shape[0] == 0
        or mouse_times.shape[0] < to_target_num_timestamps_before_visible
    ):
        return np.nan
    return mouse_times[-1] - mouse_times[to_target_num_timestamps_before_visible]


def _distance(mouse_positions: np.ndarray) -> float:
    """
    The euclidean point-to-point distance travelled by the cursor.

    :param mouse_positions: The array of mouse positions
    :return: The distance travelled.
    """
    dist = 0
    for i in range(mouse_positions.shape[0] - 1):
        dist += xydist(mouse_positions[i + 1], mouse_positions[i])
    return dist


def _rmse(mouse_positions: np.ndarray, target_position: np.ndarray) -> float:
    """
    The Root Mean Square Error (RMSE) of the perpendicular distance from each mouse point
    to the straight line that intersects the initial mouse location and the target.

    See: https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points.

    :param mouse_positions: The array of mouse positions
    :param target: The x,y coordinates of the target
    :return: The RMSE of the distance from the ideal trajectoty
    """
    if mouse_positions.shape[0] <= 1:
        return np.nan
    # use first mouse position as origin point, so exclude it from RMSE measure
    x1, y1 = mouse_positions[0]
    x2, y2 = target_position
    norm = np.power(x2 - x1, 2) + np.power(y2 - y1, 2)
    norm *= mouse_positions.shape[0] - 1
    sum_of_squares = 0
    for x0, y0 in mouse_positions[1:]:
        sum_of_squares += np.power((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1), 2)
    return np.sqrt(sum_of_squares / norm)
